/**********
This library is free software; you can redistribute it and/or modify it under
the terms of the GNU Lesser General Public License as published by the
Free Software Foundation; either version 3 of the License, or (at your
option) any later version. (See <http://www.gnu.org/copyleft/lesser.html>.)

This library is distributed in the hope that it will be useful, but WITHOUT
ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for
more details.

You should have received a copy of the GNU Lesser General Public License
along with this library; if not, write to the Free Software Foundation, Inc.,
51 Franklin Street, Fifth Floor, Boston, MA 02110-1301  USA
**********/
// "liveMedia"
// Copyright (c) 1996-2017 Live Networks, Inc.  All rights reserved.
// A file source that is a plain byte stream (rather than frames)
// Implementation

#include "ByteStreamFileSource.hh"
#include "InputFile.hh"
#include "GroupsockHelper.hh"

////////// ByteStreamFileSource //////////

// -------- start of andy2301Mod ---------------//
////// the following variables are used for monitorChanges ////////
unsigned long sleep_msec = 500;          // read a buf from each chunk of /tmp/view, sleep for sleep_msec, and read another buf to compare which chunk changed
unsigned long start_offset = 0;          // the starting file offset to read from /tmp/view
unsigned long repeat_offset = 1050000;   // we rewind the file to the starting file offset when it reaches this repeat_offset. Basically we only read [start, repeat].
unsigned long cmpsize = 10240;           // the number of bytes of data to read from each chunk of /tmp/view
unsigned int num_chunks = 7;             // we split the [start, repeat] offset range of /tmp/view into num_chunks chunks, in order to determine which chunk has changed
int adjustIdx = -1;                      // if we detect that chunk[5] changed, we may want to start reading from chunk[5 + adjustIdx]

//int max_read_size = 150000;
int reads_per_mon = 0;                   // how many doReadFromFile invokes (file reads) before we run each monitorChanges. 0 = disable monitorChanges
int num_reads = 0;                       // increments by 1 for each doReadFromFile

long offset = start_offset;              // offset tracks the current cursor of file
int verbose = 0;                         // print out more info if set to 1

//
// Monitor which part of file /tmp/view just got changed, so that we can read from there to get the latest video stream.
// The purpose is to make sure we stream the latest video segments rather than possibly 10 seconds lagging behing (file /tmp/view stores roughly 10 sec of video)
// 1. divide file offset range [start_offset, repeat_offset] of /tmp/view into num_chunks chunks (logically)
// 2. take a snapshot of <cmpsize> bytes of data from each chunk [chunkStartOfs, chunkStartOfs+cmpsize]
// 3. sleep for sleep_msec mili-seconds
// 4. take another snapshot of <cmpsize> bytes of data from each chunk [chunkStartOfs, chunkStartOfs+cmpsize]
// 5. for each chunk, compare the two snapshots for changes. Mark the chunk as changed if any.
// 6. return the chunkStartOfs offset of the chunk that just got changed if adjustIdx=0, otherwise, return the offset of chunk id (changedChunkIdx + adjustIdx).
//    If adjustIdx = -1, return the offset of the prior chunk. This is the default. The rationale is if we read the current changing chunk, we might get fragmented frame/data.
//    If adjustIdx = -2, return the offset of the prior chunk. This is safer to avoid fragmented frame/data but may have increased laggings.
//    If adjustIdx = 1, return the offset of the next chunk. Setting it to 1 is probably just for experimental purpose.
//
long ByteStreamFileSource::monitorChanges(unsigned long sleep_msec, unsigned long start_offset, 
             unsigned long repeat_offset, unsigned int num_chunks, unsigned long cmpsize, int adjustIdx)
{
  unsigned long startAtLen = start_offset;
  unsigned long repeatAtLen = repeat_offset;
  int numChunks = num_chunks;
  int chunkReadBufSize = cmpsize;

  unsigned long chunkLen = (repeatAtLen - startAtLen) / numChunks; // 150000, 150K
  unsigned long chunkStartOfs[numChunks];         // the starting offset of each chunk
  uint8_t bufPrev[numChunks][chunkReadBufSize];   // store the snapshot data for each chunk
  uint8_t bufCurr[numChunks][chunkReadBufSize];   // store the 2nd snapshot data for each chunk
  int bufChanged[numChunks];
  int i=0;

  // take a snapshot for each chunk
  for (i=0; i<numChunks; i++) {
    // initialize the bufChanged flags and starting offset for each chunk
    bufChanged[i] = 0;
    chunkStartOfs[i] = startAtLen + chunkLen * i;

    // special treatment for the first chunk
    long startofs = chunkStartOfs[i];
    if (startofs == 0) startofs = 65536; // somehow the first 10k bytes of the first chunk is always changed (probably file header), 
                                         // so we take a snapshot from offset 64KB instead.

    // take a snapshot
    fseek(fFid, startofs, SEEK_SET);
    if (fread(bufPrev[i], 1, chunkReadBufSize, fFid) <= 0) {
      fprintf(stderr, "error reading %d bytes at offset %d\n", chunkReadBufSize, chunkStartOfs[i]);
    }
  }

  // sleep for sleep_msec mili-seconds
  usleep(sleep_msec * 1000);

  // take another snapshot for each chunk
  for (i=0; i<numChunks; i++) {
    // special treatment for the first chunk
    long startofs = chunkStartOfs[i];
    if (startofs == 0) startofs = 65536; // somehow the first 10k bytes of the first chunk is always changed (probably file header), 
                                         // so we take a snapshot from offset 64KB instead.

    // take a snapshot
    fseek(fFid, startofs, SEEK_SET);
    if (fread(bufCurr[i], 1, chunkReadBufSize, fFid) <= 0) {
      fprintf(stderr, "error reading %d bytes at offset %d\n", chunkReadBufSize, chunkStartOfs[i]);
    }
  }

  // for each chunk, compare the snapshots to see if any data changed
  char str[200], *p = str;
  int strlen = 0;
  long changedOffset = -1, changedIdx = -1;
  for (i=0; i<numChunks; i++) {
    if (memcmp(bufPrev[i], bufCurr[i], chunkReadBufSize) == 0) {
      strlen += sprintf(p+strlen, "%ld\t", chunkStartOfs[i]);
      //printf("%ld\t", chunkStartOfs[i]);
    } else {
      bufChanged[i] = 1;
      strlen += sprintf(p+strlen, "%ld+\t", chunkStartOfs[i]);
      //printf("%ld+\t", chunkStartOfs[i]);
    }
  }

  // find the offset of the chunk to return, after applying adjustIdx
  for (i=0; i<numChunks; i++) {
      int nextChunk = (i+1)%numChunks;
      //if (nextChunk == 0) nextChunk = 1;

      // there might be two or more chunks got changed. We identify the last one, and then apply adjustIdx
      if (bufChanged[i] == 1 && bufChanged[nextChunk] != 1) {
        // apply adjustIdx 
        changedIdx = (i + adjustIdx + numChunks) % numChunks; // adjustIdx=1, choose the next chunk; adjustIdx=-1, choose the prior chunk
        changedOffset = chunkStartOfs[changedIdx];
        break;
      }
  }
  
  if (verbose > 0)
    fprintf(stderr, "monitorChanges: %s [%d, %ld], %d\n", p, changedIdx, changedOffset, adjustIdx);

  return changedOffset;
}
// -------- end of andy2301Mod ---------------//

ByteStreamFileSource*
ByteStreamFileSource::createNew(UsageEnvironment& env, char const* fileName,
				unsigned preferredFrameSize,
				unsigned playTimePerFrame) {
  FILE* fid = OpenInputFile(env, fileName);
  if (fid == NULL) return NULL;

  ByteStreamFileSource* newSource
    = new ByteStreamFileSource(env, fid, preferredFrameSize, playTimePerFrame);
  newSource->fFileSize = GetFileSize(fileName, fid);

  return newSource;
}

ByteStreamFileSource*
ByteStreamFileSource::createNew(UsageEnvironment& env, FILE* fid,
				unsigned preferredFrameSize,
				unsigned playTimePerFrame) {
  if (fid == NULL) return NULL;

  ByteStreamFileSource* newSource = new ByteStreamFileSource(env, fid, preferredFrameSize, playTimePerFrame);
  newSource->fFileSize = GetFileSize(NULL, fid);

  return newSource;
}

void ByteStreamFileSource::seekToByteAbsolute(u_int64_t byteNumber, u_int64_t numBytesToStream) {
  SeekFile64(fFid, (int64_t)byteNumber, SEEK_SET);

  fNumBytesToStream = numBytesToStream;
  fLimitNumBytesToStream = fNumBytesToStream > 0;
}

void ByteStreamFileSource::seekToByteRelative(int64_t offset, u_int64_t numBytesToStream) {
  SeekFile64(fFid, offset, SEEK_CUR);

  fNumBytesToStream = numBytesToStream;
  fLimitNumBytesToStream = fNumBytesToStream > 0;
}

void ByteStreamFileSource::seekToEnd() {
  SeekFile64(fFid, 0, SEEK_END);
}

ByteStreamFileSource::ByteStreamFileSource(UsageEnvironment& env, FILE* fid,
					   unsigned preferredFrameSize,
					   unsigned playTimePerFrame)
  : FramedFileSource(env, fid), fFileSize(0), fPreferredFrameSize(preferredFrameSize),
    fPlayTimePerFrame(playTimePerFrame), fLastPlayTime(0),
    fHaveStartedReading(False), fLimitNumBytesToStream(False), fNumBytesToStream(0) {
#ifndef READ_FROM_FILES_SYNCHRONOUSLY
  makeSocketNonBlocking(fileno(fFid));
#endif

  // Test whether the file is seekable
  fFidIsSeekable = FileIsSeekable(fFid);

  // -------- start of andy2301Mod ---------------//
  ///////// retrieve env variable and set values for sleep_msec, start_offset, repeat_offset, cmpsize, num_chunks and reads_per_mon ////////
  char *str = NULL;
  char printStr[200];
  int nm=0, len=0, numEnv=0;

  // get_offset setting
  str = getenv("startOFS");
  if (str && sscanf (str, "%i", &nm) == 1 && nm >= 0) {
    start_offset = nm;
    len += sprintf(printStr+len, "startOFS=%d ", start_offset);
    numEnv ++;
  }

  // repeat_offset setting
  str = getenv("repeatOFS");
  if (str && sscanf (str, "%i", &nm) == 1 && nm > 0) {
    repeat_offset = nm;
    len += sprintf(printStr+len, "repeatOFS=%d ", repeat_offset);
    numEnv ++;
  }

  // reads_per_mon, how many doReadFromFile before running a monitorChanges. 
  // This is also serves switch of enabling monitorChanges. 
  // default=0, no monitoChanges invocations. My tests show that invoking monitorchanges could bring stabilty issue to the client (random client exit)
  // if larger than 0 (e.g. 14), monitorChanges will be invoked once after every 14 invocations of doReadFromFile
  str = getenv("readsPMON");
  if (str && sscanf (str, "%i", &nm) == 1 && nm >= -1) {
    reads_per_mon = nm;
    len += sprintf(printStr+len, "readsPMON=%d ", reads_per_mon);
    numEnv ++;
  }

  // how many chunks to divide the file into
  str = getenv("numCHUNK");
  if (str && sscanf (str, "%i", &nm) == 1 && nm > 0) {
    num_chunks = nm;
    len += sprintf(printStr+len, "numCHUNK=%d ", num_chunks);
    numEnv ++;
  }

  // cmpsize setting, the size of snapshot data. Increase it to get better chance of capturing the change
  str = getenv("cmpSZ");
  if (str && sscanf (str, "%i", &nm) == 1 && nm > 0) {
    cmpsize = nm;
    len += sprintf(printStr+len, "cmpSZ=%d ", cmpsize);
    numEnv ++;
  }

  // sleep_msec setting
  str = getenv("sleepMS");
  if (str && sscanf (str, "%i", &nm) == 1 && nm > 0) {
    sleep_msec = nm;
    len += sprintf(printStr+len, "sleepMS=%d ", sleep_msec);
    numEnv ++;
  }

  // adjustIdx setting
  str = getenv("adjIDX");
  if (str && sscanf (str, "%i", &nm) == 1) {
    adjustIdx = nm;
    len += sprintf(printStr+len, "adjIDX=%d ", adjustIdx);
    numEnv ++;
  }

  // verbose level, print logging to console if verbose > 0
  str = getenv("VERB");
  if (str && sscanf (str, "%i", &nm) == 1) {
    verbose = nm;
    len += sprintf(printStr+len, "VERB=%d ", verbose);
    numEnv ++;
  }

  if (numEnv > 0) printf("ENV: %s\n", printStr);
  
  // we run monitorChanges to detect which offset in /tmp/view to stream from. Before any doReadFromFile request comes in for a session.
  // this is invoked at ByteStreamFileSource initalization stage (apparently this is initialized for each VLC client connection)
  // this is enabled by default (reads_per_mon=0), or when readsPMON>0 (periodic monitorChanges is set)
  // if env readsPMON is set to negative (allowing -1), we skip this initial run of monitorChanges. 
  if (reads_per_mon >= 0) {
    long latestChangeOffset = monitorChanges(1500, start_offset, repeat_offset, num_chunks, 30720, adjustIdx);
    if (latestChangeOffset > 0) {
      fseek(fFid, latestChangeOffset, SEEK_SET);
      offset = latestChangeOffset;
      if (verbose > 0)
        fprintf (stderr, "++ rewind to offset %ld with initial monitorChanges.\n", offset);
    }
  }
  // -------- end of andy2301Mod ---------------//
}

ByteStreamFileSource::~ByteStreamFileSource() {
  if (fFid == NULL) return;

#ifndef READ_FROM_FILES_SYNCHRONOUSLY
  envir().taskScheduler().turnOffBackgroundReadHandling(fileno(fFid));
#endif

  CloseInputFile(fFid);
}

void ByteStreamFileSource::doGetNextFrame() {
  if (feof(fFid) || ferror(fFid) || (fLimitNumBytesToStream && fNumBytesToStream == 0) || offset >= (repeat_offset-512)) {
    // andy2301Mod: reset to the beginning of this file
    if (verbose > 0)
      fprintf (stderr, "-- rewind to begining of file, prev offset=%ld, new offset=%ld\n", offset, start_offset);
    fseek(fFid, start_offset, SEEK_SET);
    offset = start_offset;
    /*
    handleClosure();
    return;
    */
  }

  // fprintf (stderr, "doGetNextFrame(): fHaveStartedReading=%d\n", fHaveStartedReading);
  // output always like this: doGetNextFrame(): fHaveStartedReading=0

#ifdef READ_FROM_FILES_SYNCHRONOUSLY
  doReadFromFile();
#else
  if (!fHaveStartedReading) {
    // Await readable data from the file:
    envir().taskScheduler().turnOnBackgroundReadHandling(fileno(fFid),
	       (TaskScheduler::BackgroundHandlerProc*)&fileReadableHandler, this);
    fHaveStartedReading = True;
  }
#endif
}

void ByteStreamFileSource::doStopGettingFrames() {
  envir().taskScheduler().unscheduleDelayedTask(nextTask());
#ifndef READ_FROM_FILES_SYNCHRONOUSLY
  envir().taskScheduler().turnOffBackgroundReadHandling(fileno(fFid));
  fHaveStartedReading = False;
#endif
}

void ByteStreamFileSource::fileReadableHandler(ByteStreamFileSource* source, int /*mask*/) {
  if (!source->isCurrentlyAwaitingData()) {
    // andy2301Mod: debug info: this return didn't cause the client exit. Now comment it out.
    //fprintf (stderr, "source not awaiting data, offset=%ld\n", offset);
    // every other request ends up here, but the other request goes through to doReadFromFile: source not awaiting data, offset=899995 
    source->doStopGettingFrames(); // we're not ready for the data yet
    return;
  }
  source->doReadFromFile();
}

void ByteStreamFileSource::doReadFromFile() {
  // -------- start of andy2301Mod ---------------//
  /*
  if (num_reads == 0) {
    fseek(fFid, start_offset, SEEK_SET);
    offset = start_offset;
  }
  */
  if (reads_per_mon>0 && num_reads % reads_per_mon == 0) {
    int sleep_msec1 = sleep_msec, cmpsize1 = cmpsize;
    /*
    if (num_reads == 0) {
      sleep_msec1 = 1500; // bump from 500 to 1500 to make sure we can capture which chunk of /tmp/view get changed
      cmpsize1 = 30720; // bump from 10240/10kB to 30kB
    }
    */
    long latestChangeOffset = monitorChanges(sleep_msec1, start_offset, repeat_offset, num_chunks, cmpsize1, adjustIdx);
    if (latestChangeOffset > 0) {
      fseek(fFid, latestChangeOffset, SEEK_SET);
      offset = latestChangeOffset;
      if (verbose > 0)
        fprintf (stderr, "++ rewind to offset %ld, num_reads=%d, reads_per_mon=%d\n", offset, num_reads, reads_per_mon);
    }
  }
  num_reads++;
  // -------- end of andy2301Mod ---------------//

  // Try to read as many bytes as will fit in the buffer provided (or "fPreferredFrameSize" if less)
  if (fLimitNumBytesToStream && fNumBytesToStream < (u_int64_t)fMaxSize) {
    fMaxSize = (unsigned)fNumBytesToStream;
  }
  if (fPreferredFrameSize > 0 && fPreferredFrameSize < fMaxSize) {
    fMaxSize = fPreferredFrameSize;
  }
#ifdef READ_FROM_FILES_SYNCHRONOUSLY
  fFrameSize = fread(fTo, 1, fMaxSize, fFid);
#else
  if (fFidIsSeekable) {
    // -------- andy2301Mod ---------------//
    long bytesread = fMaxSize;
    if ((offset + fMaxSize) >= repeat_offset) {
      bytesread = repeat_offset - offset;
    }
    fFrameSize = fread(fTo, 1, bytesread, fFid);

    //if (fFrameSize < 149990) {
    //  fseek(fFid, offset, SEEK_SET);
    //  fFrameSize = fread(fTo, 1, bytesread, fFid);
    //}
    offset += fFrameSize;
  } else {
    // For non-seekable files (e.g., pipes), call "read()" rather than "fread()", to ensure that the read doesn't block:
    fFrameSize = read(fileno(fFid), fTo, fMaxSize);
  }
#endif
  if (fFrameSize == 0) {
    // andy2301Mod: reset to the beginning of this file
    if (verbose > 0)
      fprintf (stderr, "-- rewind to begining of file (fFrameSize=0), prev offset=%ld, new offset=%ld\n", offset, start_offset);
    fseek(fFid, start_offset, SEEK_SET);
    offset = start_offset;
    /*
    handleClosure();
    return;
    */
  }
  fNumBytesToStream -= fFrameSize;

  // Set the 'presentation time':
  if (fPlayTimePerFrame > 0 && fPreferredFrameSize > 0) {
    if (fPresentationTime.tv_sec == 0 && fPresentationTime.tv_usec == 0) {
      // This is the first frame, so use the current time:
      gettimeofday(&fPresentationTime, NULL);
    } else {
      // Increment by the play time of the previous data:
      unsigned uSeconds	= fPresentationTime.tv_usec + fLastPlayTime;
      fPresentationTime.tv_sec += uSeconds/1000000;
      fPresentationTime.tv_usec = uSeconds%1000000;
    }

    // Remember the play time of this data:
    fLastPlayTime = (fPlayTimePerFrame*fFrameSize)/fPreferredFrameSize;
    fDurationInMicroseconds = fLastPlayTime;
  } else {
    // We don't know a specific play time duration for this data,
    // so just record the current time as being the 'presentation time':
    gettimeofday(&fPresentationTime, NULL);
  }

  // -------- andy2301Mod ---------------//
  if (verbose > 0) {
    char buffer[30];
    struct timeval tv;
    gettimeofday(&tv, NULL);
    strftime(buffer,30,"%m-%d %T",localtime(&tv.tv_sec));

    /*
    fprintf (stderr, "%lu.%06u [%s]: read %d bytes, offset=%ld, fPresentationTime=%lu.%06u, fPlayTimePerFrame=%d, fPreferredFrameSize=%d, fLastPlayTime=%d, fDurationInMicroseconds=%d\n", 
      tv.tv_sec, tv.tv_usec, buffer, fFrameSize, offset, fPresentationTime.tv_sec, fPresentationTime.tv_usec,
      fPlayTimePerFrame, fPreferredFrameSize, fLastPlayTime, fDurationInMicroseconds);
      // output always like the following. fPlayTimePerFrame...etc are always 0
      // 1492122646.331674 [04-13-2017 22:30:46]: read 149999 bytes, offset=1049994, 
      // fPresentationTime=1492122646.331668, fPlayTimePerFrame=0, fPreferredFrameSize=0, fLastPlayTime=0, fDurationInMicroseconds=0
    */

    printf ("%lu.%03u [%s]: readbytes=%d, offset=%ld\n", tv.tv_sec, tv.tv_usec/1000, buffer, fFrameSize, offset);
  }

  // Inform the reader that he has data:
#ifdef READ_FROM_FILES_SYNCHRONOUSLY
  // To avoid possible infinite recursion, we need to return to the event loop to do this:
  nextTask() = envir().taskScheduler().scheduleDelayedTask(0,
				(TaskFunc*)FramedSource::afterGetting, this);
#else
  // Because the file read was done from the event loop, we can call the
  // 'after getting' function directly, without risk of infinite recursion:
  FramedSource::afterGetting(this);
#endif
}
